
/******************************************************************************
* MODULE     : proof.cc
* DESCRIPTION: proofs of type conversions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <proof.hh>
#include <quantor.hh>
#include <match.hh>
#include <single.hh>
namespace mmx {

/******************************************************************************
* Merging different proofs together (while merging the instantiations)
******************************************************************************/

static list
merge_dependencies (proof p, list l) {
  if (nil (l)) return list ();
  list  pair= as_list (car (l));
  table inst= copy (as_table (pair[0]));
  list  deps= as_list (pair[1]);
  list  tail= merge_dependencies (p, cdr (l));
  if (!merge_matches (inst, p->inst)) return tail;
  return cons (list (inst, cons (p, deps)), tail);
}

static list
merge_dependencies (list deps) {
  if (nil (deps)) return list (generic (list (table (), list ())));
  proof p= as_proof (car (deps));
  list l1, l2, r;
  if (is_ambiguous (p)) l1= p->deps;
  else l1= list (p);
  l2= merge_dependencies (cdr (deps));
  while (!nil (l1)) {
    r= merge_dependencies (as_proof (car (l1)), l2) * r;
    l1= cdr (l1);
  }
  return r;
}

/******************************************************************************
* Constructors
******************************************************************************/

static list
construct_proof_sub (generic f, generic rule, list l) {
  if (nil (l)) return list ();
  list  pair= as_list (car (l));
  table inst= as_table (pair[0]);
  list  deps= as_list (pair[1]);
  list  tail= construct_proof_sub (f, rule, cdr (l));
  return cons (proof (f, rule, deps, inst), tail);
}

static proof
construct_proof (generic f, generic rule, list deps) {
  list l= merge_dependencies (deps);
  if (nil (l)) return proof ();
  l= construct_proof_sub (f, rule, l);
  return ambiguous_proof (f, l);
}

proof::proof (generic from, generic to, table t):
  rep (new proof_rep (implies (from, to), symbol ("match"), list (), t)) {}

proof::proof (generic from, generic to, string rule, list deps):
  rep (NULL)
{
  if ((N(deps)==1) && (!is_ambiguous (as_proof (deps[0])))) {
    proof p (implies (from, to), symbol (rule), deps, as_proof(deps[0])->inst);
    rep= p.rep;
  }
  else {
    proof p= construct_proof (implies (from, to), symbol (rule), deps);
    if (!nil (p)) rep= p.rep;
  }
}

proof::proof (generic from, generic to, generic rule, list deps): rep (NULL) {
  if ((N(deps)==1) && (!is_ambiguous (as_proof (deps[0])))) {
    proof p (implies (from, to), rule, deps, as_proof(deps[0])->inst);
    rep= p.rep;
  }
  else {
    proof p= construct_proof (implies (from, to), rule, deps);
    if (!nil (p)) rep= p.rep;
  }
}

proof
ambiguous_proof (generic what, list sols) {
  if (nil (sols)) return proof ();
  if (nil (cdr (sols))) return as_proof (car (sols));
  return proof (what, symbol ("ambiguous"), sols, table ());
}

/******************************************************************************
* Required proofs for generic structures
******************************************************************************/

int
proof_rep::Type () {
  return PROOF_TYPE;
}

int
proof_rep::Hash () {
  return hash (list (f, rule, deps));
}

bool
proof_rep::Equal (generic g) {
  if (gentype (g) != PROOF_TYPE) return false;
  return proof (this) == as_proof (g);
}

generic
proof_rep::Copy () {
  return copy (proof (this));
}

int
proof_rep::Arity () {
  return 0;
}

void
proof_rep::Print (ostream& out) {
  out << proof (this);
}

/******************************************************************************
* Standard routines for proofs
******************************************************************************/

ostream&
operator << (ostream& out, proof p) {
  if (nil (p)) return out << "failed" << lf;
  if (N (p->deps) == 0);
  else if (N (p->deps) == 1)
    out << as_proof (p->deps[0]);
  else {
    int  i= 1;
    list l= p->deps;
    out << indent;
    while (!nil (l)) {
      out << list (p->rule, single (i)) << lf;
      out << indent << as_proof (car (l)) << unindent;
      i++;
      l= cdr (l);
    }
    out << unindent;
  }
  return out << p->f << ",  [" << p->rule << "]" << lf;
}

proof
copy (proof p) {
  return proof (p->f, p->rule, p->deps, p->inst);
}

/******************************************************************************
* Substitution
******************************************************************************/

static table
substitute_inst (table inst, generic what, generic by) {
  table r;
  list l= inst->entries();
  while (!nil (l)) {
    generic var= car (l)[0];
    generic val= car (l)[1];
    if (var != what) r (var)= substitute (val, what, by);
    l= cdr (l);
  }
  return r;
}

static list
substitute_sub (list l, generic what, generic by) {
  if (nil (l)) return l;
  return cons (substitute (as_proof (car (l)), what, by),
	       substitute_sub (cdr (l), what, by));
}

proof
substitute (proof p, generic what, generic by) {
  return proof (substitute (p->f, what, by),
		substitute (p->rule, what, by),
		substitute_sub (p->deps, what, by),
		substitute_inst (p->inst, what, by));
}

/******************************************************************************
* Requantification of free and witness variables in proofs
******************************************************************************/

proof
requantify (proof p) {
  generic var= highest_in (p->f);
  if (nil (var)) return p;
  if (is_free_var (var)) {
    generic bv= bound_var (as_quantor(var)->name);
    generic wv= witness_var (as_quantor(var)->name);
    generic bf= substitute (p->f, var, bv);
    proof   q = substitute (p, var, wv);
    return requantify (proof (list (MMX_FORALL, bv, bf),
			      list (symbol ("=>forall"), wv),
			      list (q), p->inst));
  }
  if (is_witness_var (var)) {
    generic bv= bound_var (as_quantor(var)->name);
    generic bf= substitute (p->f, var, bv);
    return requantify (proof (list (MMX_EXISTS, bv, bf),
			      list (symbol ("=>exists"), var),
			      list (p), p->inst));
  }
  return p;
}

list
requantify (list l) {
  if (nil (l)) return l;
  proof p= as_proof (car (l));
  return cons (requantify (p), requantify (cdr (l)));
}

} // namespace mmx
